
llderivgrowth = function(obs.growthcurves, obs.times, step = 1,df=6)
{
  ######################################################################################################
  # "llderivgrowth" estimates growth curves and their 1st derivative by using the local linear regression 
  # "Responses" vector of length n containing the scalar responses Y1,...,Yn 
  # "obs.growthcurves" list containing k observed growth data: 
  #          obs.growthcurves[[1]]: X_1(t_{1,1}),...,X_1(t_{1,p_1})
  #          obs.growthcurves[[2]]: X_2(t_{2,1}),...,X_2(t_{2,p_2})
  #          .................................................
  #          obs.growthcurves[[k]]: X_k(t_{k,1}),...,X_k(t_{k,p_k})
  # "obs.times" list containing the grid of measurements (times) for each observed growth data:
  #          obs.times[[1]]: t_{1,1},...,t_{1,p_1}
  #          obs.times[[2]]: t_{2,1},...,t_{2,p_2}
  #          .................................
  #          obs.times[[k]]: t_{k,1},...,t_{k,p_k}
  # "step" is the time unit allowing to build an equally-spaced grid points over which the growth curves and their 1st derivatives 
  #          are to be estimated; "step" must match the realtime unit used for the varying temperature sequence 
  #          (by default, "step" is set to 1)
  ##################################################################################################
  # Returns a list containing
  # "$Eval" gives the equally-spaced grid points over which the 1st derivatives are to be estimated
  # "$estimated.deriv.growth": list containing the estimated 1st derivatives of growth curves
  # "$estimated.growth": list containing the estimated growth curves
  ##################################################################################################
  ## require "KernSmooth" and "sm" packages 
  require(KernSmooth)
  require(sm)
  K = length(obs.growthcurves)
  estimated.growthcurves = list()
  estimated.deriv.growthcurves = list()
  Eval = list()
  for(kk in 1:K){		
    MinMax = range(obs.times[[kk]])
    Eval[[kk]] = seq(MinMax[1], MinMax[2] + step, by = step)
    Range.grid = range(Eval[[kk]])
    hcv = h.select(obs.times[[kk]], obs.growthcurves[[kk]],df=df)
    estimated.growthcurves[[kk]] = locpoly(obs.times[[kk]], obs.growthcurves[[kk]], drv = 0L, degree = 1, kernel = "normal", bandwidth = hcv, gridsize = length(Eval[[kk]]), bwdisc = 25, 
                                           range.x = Range.grid, binned = FALSE, truncate = TRUE)$y
    estimated.deriv.growthcurves[[kk]] = locpoly(obs.times[[kk]], obs.growthcurves[[kk]], drv = 1L, degree = 1, kernel = "normal", bandwidth = hcv, gridsize = length(Eval[[kk]]), bwdisc = 25, 
                                                 range.x = Range.grid, binned = FALSE, truncate = TRUE)$y
  }
  return(list(Eval = Eval, estimated.deriv.growthcurves = estimated.deriv.growthcurves, estimated.growthcurves = estimated.growthcurves))
}


resulting.varying.temp.growth.profile = function(results.pmi, benchmark.eval, benchmark.growthcurves, Benchmark.temp, zoom)
{
  ######################################################################################################
  # "resulting.varying.temp.growth.profile" displays the estimated temperature-dependent growth curve 
  #######################################################################################################
  # "results.pmi" list containing the results provided by the "pmi_estimation" procedure.
  # "benchmark.eval" list containing the sorted grid points at which each benchmark growth curves are evaluated:
  #          benchmark.eval[[1]]: s_{1,1}, s_{1,2},..., s_{1,p_1} 
  #          benchmark.eval[[2]]: s_{2,1}, s_{2,2},..., s_{2,p_2} 
  #          ....................................................
  #          benchmark.eval[[K]]: s_{K,1}, s_{K,2},..., s_{K,p_K} 
  # "benchmark.growthcurves" list containing K benchmark growth curves:
  #          benchmark.growthcurves[[1]]: X_1(s_{1,1}), X_1(s_{1,2}),..., X_1(s_{1,p_1}) 
  #          benchmark.growthcurves[[2]]: X_2(s_{2,1}), X_2(s_{2,2}),..., X_2(s_{2,p_2}) 
  #          .....................................................................
  #          benchmark.growthcurves[[K]]: X_K(s_{K,1}), X_K(s_{K,2}),..., X_K(s_{K,p_K}) 
  # "Benchmark.temp" vector of benchmark temperatures
  # "zoom" logical value; if zoom = F, the procedure displays the estimated 
  #      temperature-dependent growth profile superimposed on the whole benchmark growth curves; 
  #      if zoom = T, the benchmark growth curves are truncated in order to focus on the estimated 
  #      temperature-dependent growth profile
  ##################################################################################################
  Maxtimes = sapply(benchmark.eval, max)
  maxtime =  max(Maxtimes)
  mintime = min(sapply(benchmark.eval, min))
  K = length(benchmark.eval)
  Varying.temp.time.progression = results.pmi$Times.progression
  Varying.temp.growth.shape = results.pmi$Growth.shape
  if(zoom){
    xlim = range(Varying.temp.time.progression)
    ylim = range(Varying.temp.growth.shape)
    title = paste("Zoom on the estimated growth curve (pmi = ", results.pmi$pmi, ")")
  }else{ 
    xlim = c(mintime, maxtime)
    ylim = range(cbind(sapply(benchmark.growthcurves, range), sapply(Varying.temp.growth.shape, range)))
    title = paste("5) Global view on the estimated growth curve (pmi = ", results.pmi$pmi, ")")
  }
  par(mar = c(5, 5, 5, 1))
  plot(Varying.temp.time.progression, Varying.temp.growth.shape, type = "l", xlab = "time (hours)", ylab = "lengths", 
       xlim = xlim, ylim = ylim, lwd = 2)
  abline(v = max(Varying.temp.time.progression), lty = 2)
  for(kk in 1:K){
    par(new = T)
    plot(benchmark.eval[[kk]], benchmark.growthcurves[[kk]], type = "l", xlab = "", ylab = "", xlim = xlim, ylim = ylim, lwd = 0.5, main = title)
  }
}

forecasting_crime_scene_temperatures = function(Temp.cs, Temp.ws, spar = 10)
{
  ######################################################################################################
  # "forecasting_crime_scene_temperatures" provides nonparametric forecast of the crime scene temperature 
  #######################################################################################################
  # "Temp.cs" vector of temperature observed at the crime scene from latest to oldest (after body discovery)
  # "Temp.ws" vector of temperature from the weather station from latest to oldest
  ##################################################################################################
  ## require "np" package
  
  require(np)
  p = length(Temp.ws)
  n<-length(Temp.cs)
  Temp.ws.smooth = ksmooth(1:p, Temp.ws, bandwidth = spar, x.points = 1:p)$y	
  Temp.ws.residual = Temp.ws - Temp.ws.smooth
  Temp.cs.residual = Temp.cs - Temp.ws.smooth[1:n]
  bw = npregbw(xdat = Temp.ws.residual[2:(n+1)], ydat = Temp.cs.residual)
  forecast = npreg(bw, exdat = Temp.ws.residual[-(1:(n+1))])$mean
  return(forecast + Temp.ws.smooth[-(1:(n+1))])
}
#####################################################################
#####################################################################
#####################################################################
registered.growthcurves <- function(benchmark.eval, benchmark.growthcurves, nknots, norder, landmark0 = 1030)
{
  ######################################################################################################
  # "registered.growthcurves" achieves a registration by setting landmarks (use of the "fda" R package).  
  # Two sets of curves are computed: the aligned growth profiles and the corresponding warping functions.
  #######################################################################################################
  # "benchmark.eval" list containing the sorted grid points at which each benchmark growth curves are evaluated:
  #          benchmark.eval[[1]]: s_{1,1}, s_{1,2},..., s_{1,p_1} 
  #          benchmark.eval[[2]]: s_{2,1}, s_{2,2},..., s_{2,p_1} 
  #          ....................................................
  #          benchmark.eval[[K]]: s_{K,1}, s_{K,2},..., s_{K,p_1} 
  # "benchmark.growthcurves" list containing K benchmark growth curves:
  #          benchmark.growthcurves[[1]]: X_1(s_{1,1}), X_1(s_{1,2}),..., X_1(s_{1,p_1}) 
  #          benchmark.growthcurves[[2]]: X_2(s_{2,1}), X_2(s_{2,2}),..., X_2(s_{2,p_1}) 
  #          .....................................................................
  #          benchmark.growthcurves[[K]]: X_K(s_{K,1}), X_K(s_{K,2}),..., X_K(s_{K,p_K}) 
  # "nknots" number of knots of the used B-splines basis for approximating growth curves (default is 50)
  # "norder" order of the used B-splines basis for approximating growth curves (default is 4 - cubic spline)
  # "landmark0" set the value locating the place where the maxima are aligned
  ##################################################################################################
  # Returns a list containing
  # "$ALIGNED.GROWTH.CURVES": (p * K)-matrix contained the K aligned benchmark growth curves (column by column)
  # "$WARPED.TIME": (p * K)-matrix contained the K warping functions (column by column) 
  # "$Eval.point": vector contained the grid points at which the aligned benchmark growth curves 
  #                (and the warping functions) are computed
  # "landmark0" : the value locating the place where the maxima are aligned
  ##################################################################################################
  ## require "fda" package 
  require(fda)
  K = length(benchmark.eval)
  RANGE.GRID = sapply(benchmark.eval, range)
  mintime = min(RANGE.GRID[1, ])
  maxtime = max(RANGE.GRID[2, ])
  Lengths = sapply(benchmark.eval, length)
  maxlength = max(Lengths)
  x.eval = seq(mintime, maxtime, length = maxlength)
  # rescale the time for starting and ending at the same point for each growth curve
  time.rescaled = list()
  APPROX.GROWTHS = matrix(0, maxlength, K)
  Landmarks = matrix(NA,K,1)
  for(kk in 1:K){
    time.rescaled[[kk]] = seq(mintime, maxtime, length = Lengths[kk])
    APPROX.GROWTHS[, kk]  = approx(time.rescaled[[kk]], benchmark.growthcurves[[kk]], xout = x.eval)$y
    Landmarks[kk,1] = min(x.eval[which.max(APPROX.GROWTHS[, kk])],length(APPROX.GROWTHS[, kk])-100)
  }
  # register the rescaled growth curves by aligning maxima (excepted the first one)
  knots  = seq(mintime, maxtime, length = nknots)
  
  nbasis = length(knots) + norder - 2
  #lengthbasis = create.bspline.basis(range(x.eval), norder=norder, breaks=knots)
  #growfdPar = fdPar(lengthbasis, lambda = 1e-5)
  #lengthfd = smooth.basis(x.eval, APPROX.GROWTHS, lengthbasis)$fd
  #WfdPar = fdPar(lengthfd, lambda = 1e-3)
  #landmarkres = landmarkreg(lengthfd, Landmarks, landmark0, WfdPar=WfdPar)
  
   lengthbasis <- create.bspline.basis(rangeval=range(x.eval),
                                  norder=4, breaks=seq(min(x.eval), max(x.eval), len=length(knots)))
   lengthfd = smooth.basis(x.eval, APPROX.GROWTHS, lengthbasis)$fd
   Wfd0   <- fd(matrix(0,lengthbasis$nbasis,1),lengthbasis)
   WfdPar <- fdPar(Wfd0, lambda=1e-1)
   landmarkres = landmarkreg(lengthfd, Landmarks, landmark0, WfdPar=WfdPar)
  
  WARPED.TIME = eval.fd(x.eval, landmarkres$warpfd)
  ALIGNED.GROWTH.CURVES = eval.fd(x.eval, landmarkres$regfd)
  return(list(ALIGNED.GROWTH.CURVES = ALIGNED.GROWTH.CURVES, WARPED.TIME = WARPED.TIME, Eval.points = x.eval, landmark0 = landmark0))
}	


aligned.growth.profile.approx = function(ALIGNED.BGP, Benchmark.temp, Eval.temp, mba, n, m, h,min.temp=4,bandwidth=bandwidth)
  ######################################################################################################
# "aligned.growth.profile.approx" approximates aligned growth profiles at any temperature contained in "Eval.temp"
#######################################################################################################
# "ALIGNED.BGP" (p * K)-matrix containing (column by column) the K aligned benchmark growth curves computed by the 
#			routine "registered.growthcurves"
# "Benchmark.temp" vector giving the corresponding K benchmark temperatures
# "Eval.temp" vector of temperatures at which the aligned growth profiles are approximated
# "mba" logical value: if mba=T, a multilevel B-spline surface approximation is used (see the MBA R package)
# "n, m, h" additional optional arguments for the "mba.points" R routine of the MBA R package
##################################################################################################
# Returns a matrix (l * p)-matrix contained the approximated aligned benchmark growth curves (row by row) : 
# row 1 contains the approximated aligned benchmark growth profile for "Eval.temp[1]"
# row 2 contains the approximated aligned benchmark growth profile for "Eval.temp[2]"
# ...................................................................................
# row l contains the approximated aligned benchmark growth profile for "Eval.temp[l]" where l is the length of "Eval.temp"
##################################################################################################
{
  require(fdapace)
  K = length(Benchmark.temp)
  l.eval.temp = length(Eval.temp)
  p = ncol(ALIGNED.BGP)
  APPROX.ALIGNED.GP = matrix(0, l.eval.temp, p)
  if(mba){
    require(MBA)
    x = rep(Benchmark.temp, rep(p, K))
    y = rep(1:p, K)
    z = as.vector(t(ALIGNED.BGP))
    x.new = rep(Eval.temp, rep(p, l.eval.temp))
    y.new = rep(1:p, l.eval.temp)
    COORD.NEW = cbind(x.new, y.new)
    z.new = mba.points(cbind(x, y, z), COORD.NEW, n, m, h)$xyz.est
    APPROX.ALIGNED.GP = matrix(z.new[,3], nrow = l.eval.temp, ncol = p, byrow = T)		
  }else{
    for (i in 1:p){
    #hcv = h.select(Benchmark.temp, ALIGNED.BGP[,i],df=3)
    #APPROX.ALIGNED.GP[,i]= ksmooth( Benchmark.temp, ALIGNED.BGP[,i], kernel="box",bandwidth =2*max(diff(Benchmark.temp)),  x.points = Eval.temp)$y
    APPROX.ALIGNED.GP[,i]= Lwls1D(bw=bandwidth,kernel_type = "quar", xin=Benchmark.temp, yin=ALIGNED.BGP[,i],  xout = Eval.temp,npoly = 0)
    pos.inf<-which(Eval.temp<min.temp)  
    APPROX.ALIGNED.GP[pos.inf,i]<-0
    #pos.sup<-which(Eval.temp>=max(Benchmark.temp))  
    #APPROX.ALIGNED.GP[pos.sup,i]<-ALIGNED.BGP[which.max(Benchmark.temp),i]
    
    }
  }
  return(APPROX.ALIGNED.GP)
}

warped.time.of.aligned.growth.profile.approx = function(WT.ALIGNED.BGP, Benchmark.temp, Eval.temp, mba, n, m, h,min.temp=4 ,bandwidth)
  ######################################################################################################
# "warped.time.of.aligned.growth.profile.approx" approximates warped functions aligning growth profiles 
#		at any temperature contained in "Eval.temp"
#######################################################################################################
# "WT.ALIGNED.BGP" (p * K)-matrix containing (column by column) the K warped functions computed by the 
#			routine "registered.growthcurves"
# "Benchmark.temp" vector giving the corresponding K benchmark temperatures
# "Eval.temp" vector of temperatures at which the warped functions are approximated
# "mba" logical value: if mba=T, a multilevel B-spline surface approximation is used (see the MBA R package)
# "n, m, h" additional optional arguments for the "mba.points" R routine of the MBA R package
##################################################################################################
# Returns a matrix (l * p)-matrix contained the approximated warped functions (row by row) : 
# row 1 contains the warped function for "Eval.temp[1]"
# row 2 contains the warped function for "Eval.temp[2]"
# ...................................................................................
# row l contains the warped function for "Eval.temp[l]" where l is the length of "Eval.temp"
##################################################################################################
{
  require(fdapace)
  K = length(Benchmark.temp)
  l.eval.temp = length(Eval.temp)
  p = ncol(WT.ALIGNED.BGP)
  APPROX.WT = matrix(0, l.eval.temp, p)
  if(mba){
    require(MBA)
    x = rep(Benchmark.temp, rep(p, K))
    y = rep(1:p, K)
    z = as.vector(t(WT.ALIGNED.BGP))
    x.new = rep(Eval.temp, rep(p, l.eval.temp))
    y.new = rep(1:p, l.eval.temp)
    COORD.NEW = cbind(x.new, y.new)
    z.new = mba.points(cbind(x, y, z), COORD.NEW, n, m, h)$xyz.est
    APPROX.WT = matrix(z.new[,3], nrow = l.eval.temp, ncol = p, byrow = T)		
  }else{
    for (i in 1:p){ 
      #APPROX.WT[,i]<- ksmooth(Benchmark.temp, WT.ALIGNED.BGP[,i], kernel="box",bandwidth =2*max(diff(Benchmark.temp)),  x.points = Eval.temp)$y
      APPROX.WT[,i]= Lwls1D(bw=bandwidth,kernel_type = "quar", xin=Benchmark.temp, yin=WT.ALIGNED.BGP[,i],  xout = Eval.temp,npoly = 0)
      
      pos.inf<-which(Eval.temp<min.temp)  
      APPROX.WT[pos.inf]<-0
      #pos.sup<-which(Eval.temp>=max(Benchmark.temp))  
      #APPROX.WT[pos.sup]<-WT.ALIGNED.BGP[which.max(Benchmark.temp),i]
      
    }
  }
  return(APPROX.WT)
}

approx.growth.profile = function(APPROX.ALIGNED.GP, APPROX.WARPED.TIME, Approx.maxtime,Maxtime)
  ######################################################################################################
# "approx.growth.profile" approximates growth profiles at any temperature contained in "Eval.temp"
#######################################################################################################
# "APPROX.ALIGNED.GP" (l * p)-matrix contained (row by row) the approximated aligned growth curves computed by the 
# 	routine "aligned.growth.profile.approx":
# 		row 1 contains the approximated aligned benchmark growth profile for "Eval.temp[1]"
# 		row 2 contains the approximated aligned benchmark growth profile for "Eval.temp[2]"
# 		...................................................................................
# 		row l contains the approximated aligned benchmark growth profile for "Eval.temp[l]" 
#		(where l is the length of "Eval.temp")
# "APPROX.ALIGNED.GP" (l * p)-matrix contained (row by row) the approximated warped functions computed by the 
# 	routine "warped.time.of.aligned.growth.profile.approx":
# 		row 1 contains the approximated warped functions for "Eval.temp[1]"
# 		row 2 contains the approximated warped functions for "Eval.temp[2]"
# 		..................................................................
# 		row l contains the approximated warped functions for "Eval.temp[l]" 
#		(where l is the length of "Eval.temp")
# "Approx.maxtime" vector contained the approximated maximum growth time for each temperature in "Eval.temp"
##################################################################################################
# Returns a list containing
# "$approx.GP" list containing the approximated growth curves
# "$approx.eval" list containing the corresponding time points (0, 1, 2, ...) at which the growth profiles are evaluated
##################################################################################################
{
  K = nrow(APPROX.ALIGNED.GP)
  approx.eval = list()
  approx.GP = list()
  Approx.maxtime = round(Approx.maxtime)
  maxtime = max(Maxtime) 
  for(kk in 1:K){
    approx.eval[[kk]] = 0:Approx.maxtime[kk]
    approx.GP[[kk]] = approx(APPROX.WARPED.TIME[kk, ] * Approx.maxtime[kk] / maxtime, APPROX.ALIGNED.GP[kk, ], xout = approx.eval[[kk]], rule = 2)$y
  }
  return(list(approx.GP = approx.GP, approx.eval = approx.eval))
}

# approx.GP.atany.ct = function(benchmark.eval, benchmark.growthcurves, Benchmark.temp, Eval.temp, nknots, norder, mba)
#   ######################################################################################################
# # "approx.GP.atany.ct" approximates growth profiles at any outdoor temperature contained in "Eval.temp" from the 
# # estimated benchmark growth curves derived from the routine "llderivgrowth"; this routine calls 
# # "registered.growthcurves", "aligned.growth.profile.approx", "warped.time.of.aligned.growth.profile.approx", 
# # and "approx.growth.profile"
# #######################################################################################################
# # "benchmark.eval" list containing the sorted grid points at which each benchmark growth curves are evaluated:
# #          benchmark.eval[[1]]: s_{1,1}, s_{1,2},..., s_{1,p_1} 
# #          benchmark.eval[[2]]: s_{2,1}, s_{2,2},..., s_{2,p_2} 
# #          ....................................................
# #          benchmark.eval[[K]]: s_{K,1}, s_{K,2},..., s_{K,p_K} 
# # "benchmark.growthcurves" list containing K benchmark growth curves:
# #          benchmark.growthcurves[[1]]: X_1(s_{1,1}), X_1(s_{1,2}),..., X_1(s_{1,p_1}) 
# #          benchmark.growthcurves[[2]]: X_2(s_{2,1}), X_2(s_{2,2}),..., X_2(s_{2,p_2}) 
# #          .....................................................................
# #          benchmark.growthcurves[[K]]: X_K(s_{K,1}), X_K(s_{K,2}),..., X_K(s_{K,p_K}) 
# # "Benchmark.temp" vector containing the K benchmark temperatures
# # "Eval.temp" vector of temperatures: the 1st growth curve "approx.GP[[1]]" is the approximated growth curve at 
# #          the constant temperature "Eval.temp[1]", the 2nd at "Eval.temp[2]" and so on. 
# # "nknots" an integer specifying the number of interior knots for building the b-splines (used in the registration 
# #          stage with the "fda" R package). 
# # "norder" an integer specifying the order of b-splines (used in the registration stage with the "fda" R package). 
# #          The default of 4 gives cubic splines.
# # "mba" logical value: if mba=T, a multilevel B-spline surface approximation is used (see the MBA R package)
# ##################################################################################################
# # Returns a list containing
# # "$APPROX.GP" list containing the approximated growth curves
# # "$approx.eval" list containing the corresponding time points (0, 1, 2, ...) at which the growth profiles are evaluated
# ##################################################################################################
# {
#   # registration stage (with graphical display)
#   registered.benchmark.growth.curves = registered.growthcurves(benchmark.eval, benchmark.growthcurves, nknots, norder, landmark0 = 400)
#   ALIGNED.BENCHMARK.GROWTH.PROFILES = t(registered.benchmark.growth.curves$ALIGNED.GROWTH.CURVES)
#   WARPED.TIME = t(registered.benchmark.growth.curves$WARPED.TIME)
#   Eval.points = registered.benchmark.growth.curves$Eval.points
#   # approximating aligned growth curves at constant temperature
#   APPROX.ALIGNED.GP = aligned.growth.profile.approx(ALIGNED.BENCHMARK.GROWTH.PROFILES, Benchmark.temp, Eval.temp, mba, 50, 1, 4)
#   # approximating corresponding time warping
#   APPROX.WARPED.TIME = warped.time.of.aligned.growth.profile.approx(WARPED.TIME, Benchmark.temp, Eval.temp, mba, 50, 1, 5)
#   # built approximated growth profile at contant temperatures given in "Eval.temp"
#   Maxtime = sapply(benchmark.eval, max)
#   Approx.maxtime = approx(Benchmark.temp, Maxtime, xout = Eval.temp, rule = 2)$y
#   res.agp = approx.growth.profile(APPROX.ALIGNED.GP, APPROX.WARPED.TIME, Approx.maxtime,Maxtime)
#   return(list(ABGP = ALIGNED.BENCHMARK.GROWTH.PROFILES, WT = WARPED.TIME, AAGP = APPROX.ALIGNED.GP, AWT = APPROX.WARPED.TIME, approx.eval = res.agp$approx.eval, approx.GP = res.agp$approx.GP, landmark0 = registered.benchmark.growth.curves$landmark0))	
# }



approx.GP.reg = function(benchmark.eval,registered.benchmark.growth.curves, Benchmark.temp, Eval.temp, bandwidth= 2*max(diff(Benchmark.temp)), nknots, norder, mba,min.temp=4)
  ######################################################################################################
  # "approx.GP.reg" approximates growth profiles at any outdoor temperature contained in "Eval.temp" from the 
  # registered benchmark growth curves derived from the routine "registered.growthcurves"; # "aligned.growth.profile.approx", "warped.time.of.aligned.growth.profile.approx", 
  # and "approx.growth.profile"
  #######################################################################################################
#   registered.benchmark.growth.curves  a list with the aligned grwoth curves and warping functions at the experimental temperatures the 
#          
# "Benchmark.temp" vector containing the K benchmark temperatures
# "Eval.temp" vector of temperatures: the 1st growth curve "approx.GP[[1]]" is the approximated growth curve at 
#          the constant temperature "Eval.temp[1]", the 2nd at "Eval.temp[2]" and so on. 
# "nknots" an integer specifying the number of interior knots for building the b-splines (used in the registration 
#          stage with the "fda" R package). 
# "norder" an integer specifying the order of b-splines (used in the registration stage with the "fda" R package). 
#          The default of 4 gives cubic splines.
# "mba" logical value: if mba=T, a multilevel B-spline surface approximation is used (see the MBA R package)
##################################################################################################
# Returns a list containing
# "$APPROX.GP" list containing the approximated growth curves
# "$approx.eval" list containing the corresponding time points (0, 1, 2, ...) at which the growth profiles are evaluated
##################################################################################################
{
  require(fdapace)
  # registration stage (with graphical display)
   ALIGNED.BENCHMARK.GROWTH.PROFILES = t(registered.benchmark.growth.curves$ALIGNED.GROWTH.CURVES)
  WARPED.TIME = t(registered.benchmark.growth.curves$WARPED.TIME)
  Eval.points = registered.benchmark.growth.curves$Eval.points
  # approximating aligned growth curves at constant temperature
  min.temp=min(min.temp,min(Benchmark.temp))
  APPROX.ALIGNED.GP = aligned.growth.profile.approx(ALIGNED.BENCHMARK.GROWTH.PROFILES, Benchmark.temp, Eval.temp, mba, 50, 1, 4,min.temp = min.temp,bandwidth=bandwidth)
  # approximating corresponding time warping
  APPROX.WARPED.TIME = warped.time.of.aligned.growth.profile.approx(WARPED.TIME, Benchmark.temp, Eval.temp, mba, 50, 1, 5,min.temp = min.temp,bandwidth=bandwidth)
  # built approximated growth profile at contant temperatures given in "Eval.temp"
  Maxtime = sapply(benchmark.eval, max)
  #Approx.maxtime = approx(Benchmark.temp, Maxtime, xout = Eval.temp, rule = 2)$y
 
  Approx.maxtime =Lwls1D(bw=bandwidth,kernel_type = "quar", xin=Benchmark.temp, yin=Maxtime,  xout = Eval.temp,npoly = 0)
  res.agp = approx.growth.profile(APPROX.ALIGNED.GP, APPROX.WARPED.TIME, Approx.maxtime,Maxtime)
  return(list(ABGP = ALIGNED.BENCHMARK.GROWTH.PROFILES, WT = WARPED.TIME, AAGP = APPROX.ALIGNED.GP, AWT = APPROX.WARPED.TIME, approx.eval = res.agp$approx.eval, approx.GP = res.agp$approx.GP, landmark0 = registered.benchmark.growth.curves$landmark0))	
}


########### 



# tdgp2 = function(GrowthCurvesUpMaxima, EvalPointsUpMaxima, POS.OF.BACK.TEMP, nb.of.time, Weights)
#   ######################################################################################################
# # Estimation of the temperature-dependent growth profile
# #######################################################################################################
# # "GrowthCurvesUpMaxima" list containing K approximated growth curves (truncated after their maxima):
# #          GrowthCurvesUpMaxima[[1]]: X_1(s_{1,1}), X_1(s_{1,2}),..., X_1(s_{1,p_1}) 
# #          GrowthCurvesUpMaxima[[2]]: X_2(s_{2,1}), X_2(s_{2,2}),..., X_2(s_{2,p_2}) 
# #          .....................................................................
# #          GrowthCurvesUpMaxima[[K]]: X_K(s_{K,1}), X_K(s_{K,2}),..., X_K(s_{K,p_K}) 
# # "EvalPointsUpMaxima" list containing the corresponding sorted grid points at which the approximated growth 
# #                  (truncated after their maxima) are evaluated:
# #          EvalPointsUpMaxima[[1]]: s_{1,1}, s_{1,2},..., s_{1,p_1} 
# #          EvalPointsUpMaxima[[2]]: s_{2,1}, s_{2,2},..., s_{2,p_2} 
# #          ....................................................
# #          EvalPointsUpMaxima[[K]]: s_{K,1}, s_{K,2},..., s_{K,p_K} 
# # "Eval.temp" vector of temperatures: the 1st growth curve "GrowthCurvesUpMaxima[[1]]" is the approximated growth curve at 
# #          the constant temperature "Eval.temp[1]", the 2nd at "Eval.temp[2]" and so on. 
# # "nb.temp_vary" length of the whole outdoor temperature time serie observed/estimated before the discovery of the corpse
# # "start" scalar giving the current starting date for estimating pmi
# # "WEIGHTS" matrix containing the temperature-dependent weights built from the outdoor temperature time serie and the 
# #       benchmark temperatures
# ##################################################################################################
# # Returns a vector giving the estimated length progression
# ##################################################################################################
# {
#   Length.progression = sum(sapply(GrowthCurvesUpMaxima, function(x) x[1])[POS.OF.BACK.TEMP[1,]] * Weights)
#   K = length(GrowthCurvesUpMaxima)
#   Mean.gap = 0
#   for(kk in 1:K){
#     GrowthCurvesUpMaxima[[kk]] = GrowthCurvesUpMaxima[[kk]] - GrowthCurvesUpMaxima[[kk]][1] + Length.progression[1]
#     Mean.gap[kk] = mean(diff(GrowthCurvesUpMaxima[[kk]]))
#   }
#   backward = ncol(POS.OF.BACK.TEMP)
#   Updated.delta.length = 0
#   for(ll in 2:nb.of.time){
#     count = 0
#     for(kk in POS.OF.BACK.TEMP[ll,1:min(ll, backward)]){
#       count =  count + 1
#       Eval = EvalPointsUpMaxima[[kk]]
#       Growth = GrowthCurvesUpMaxima[[kk]]
#       ## stop growth at maximum
#       # if(Length.progression[ll - 1] > max(Growth)){
#       #   Updated.delta.length[count] = 0
#       # }else{
#       updated.time = approx(Growth, Eval, xout = Length.progression[ll - 1], rule = 2)$y
#       Updated.delta.length[count] = diff(approx(Eval, Growth, xout = c(updated.time, updated.time + 1), rule = 2)$y)				
#       #}
#     }
#     Length.progression[ll] = Length.progression[ll - 1] + sum(Updated.delta.length * Weights)
#   }
#   return(Length.progression)
# }
# 




growth_estimation= function(observed.lengths, Times_temp_vary, Temp_vary,res.agpaact, Eval.temp,  Search,time.limits=rep(length(Times_temp_vary),length(observed.lengths)),postfeeding=rep(NULL,length(observed.lengths)))

{
  ######################################################################################################
  # Estimation of the hatching instant; as by-product, the procedure furnishes the estimated 
  # temperature-dependent growth profile. 
  #######################################################################################################
  # "Observed.lengths" vector containing observed lengths of maggots taken into the corpse
  # "Times_temp_vary" vector of times at which the varying temperatures are observed/estimated. 
  # "Temp_vary" vector of observed/estimated temperature at times "Times_temp_vary". 
  # "res.agpaact" output of the approx.GP.reg function, containing constant-temperature growth curves for all
  #               temperature in Eval.temp
  #          approx.GP[[K]]: X_K(s_{K,1}), X_K(s_{K,2}),..., X_K(s_{K,p_K}) 
  # "Eval.temp" vector of temperatures: the 1st growth curve "approx.GP[[1]]" is the approximated growth curve at 
  #          the constant temperature "Eval.temp[1]", the 2nd at "Eval.temp[2]" and so on. 
  # "Search" vector giving a set (grid) of dates from which the optimal one is selected
  # "time.limit"= vector of the furthest observed times in the experimental data for each species
  # "postfeeding" logical, vector which denotes if each species was observed in the postfeeding phase. NULL if this information is not available.
  ##################################################################################################
  # Returns a list containing
  # "$Growth.shape" vector containing the estimated temperature-dependent growth shape for the optimal date
  # "$Times.progression" vector containing the corresponding time progression
  # "$start.opt" give the selected starting date
  # "$pmi" gives the estimated post-mortem instant
  ##################################################################################################
  nb.temp_vary = length(Temp_vary)
  Temp_vary.pos.in.Eval.temp = as.numeric(factor(Temp_vary, levels = Eval.temp, labels = 1:length(Eval.temp)))
  Locating.maxima.benchmark.growthcurves = vector("list",length(observed.lengths))
  EvalPoints = vector("list",length(observed.lengths))
  GrowthCurves = vector("list",length(observed.lengths))
  
  for (s in 1:length(observed.lengths)){ 
    Locating.maxima.benchmark.growthcurves[[s]] = sapply(res.agpaact[[s]]$approx.GP, which.max)
    K = length(res.agpaact[[s]]$approx.GP)
    for(kk in 1:K){
      EvalPoints[[s]][[kk]] = res.agpaact[[s]]$approx.eval[[kk]]
      GrowthCurves[[s]][[kk]] = res.agpaact[[s]]$approx.GP[[kk]]
    }
  }
  
  Lengths = vector("list",length(observed.lengths))
  Times = vector("list",length(observed.lengths))
 emp_postfeeding = vector("list",length(observed.lengths))
  for (s in 1:length(observed.lengths)){ 
    count = 0
    Lengths[[s]]<-rep(NA,length(Search))
    Times[[s]]<-rep(NA,length(Search))
    for(start in Search){
      
      nb.of.time = nb.temp_vary - start + 1
      ### estimating the shape of the temperature-dependent growth profiles
      if(start>1 & start<nb.temp_vary){ 
      POS.OF.BACK.TEMP = Temp_vary.pos.in.Eval.temp[-(1:(start-1))]
      }else if(start<=1){
        POS.OF.BACK.TEMP = Temp_vary.pos.in.Eval.temp
        start<-1
        nb.of.time = nb.temp_vary - start + 1
      }else{
        start<-nb.temp_vary
        nb.of.time = nb.temp_vary - start + 1
        POS.OF.BACK.TEMP=Temp_vary.pos.in.Eval.temp[nb.of.time]
      }
      Varying.temp.process = growth_process(GrowthCurves=GrowthCurves[[s]], EvalPoints=EvalPoints[[s]], Eval.temp, temp.series=POS.OF.BACK.TEMP, nb.of.time=nb.of.time, max.times =Locating.maxima.benchmark.growthcurves[[s]] )
      ### approximating integral over benchmark temperatures ###
      count = count + 1
      Lengths[[s]][count] = Varying.temp.process$Length.progression[nb.temp_vary - start + 1]
      Times[[s]][count] = nb.temp_vary - start
      emp_postfeeding[[s]][count]<-Varying.temp.process$postfeeding
    }
  }
  ### compute criterion for selecting the optimal parameter and time
 Criterion<-rep(0,length(Search))
  for (index in 1:length(Search)){ 
    nb.of.time = nb.temp_vary - Search[index] + 1
    active.species<-0
  for (s in 1:length(observed.lengths)){ 
    if(nb.of.time<time.limits[s]){
      if(is.null(postfeeding[[s]])||postfeeding[[s]]==emp_postfeeding[[s]][index]){
        if(!is.na(Lengths[[s]][index])){ 
      Criterion[index] = Criterion[index]+ (mean(observed.lengths[[s]]) - Lengths[[s]][index])^2
      active.species<-active.species+1
        }
      }
    }
  }
    if(active.species>0){ 
    Criterion[index]<-Criterion[index]/active.species
    }
  }
  Criterion[which(Criterion==0)]<-Inf
  
  if(Search[which.min(Criterion)]>=1 & Search[which.min(Criterion)]<=nb.temp_vary){ 
  start.opt = Search[which.min(Criterion)]
  }else if(Search[which.min(Criterion)]<1){
    start.opt<-1
  }else{
    start.opt<-nb.temp_vary
    }
  
  ### estimating the shape of the temperature-dependent growth curves
  nb.of.time = nb.temp_vary - start.opt + 1
  if(start.opt>1 & start.opt<nb.temp_vary){ 
      POS.OF.BACK.TEMP = Temp_vary.pos.in.Eval.temp[-(1:(start.opt-1))]
      }else if(start==1){
        POS.OF.BACK.TEMP = Temp_vary.pos.in.Eval.temp
       
      }else{
       POS.OF.BACK.TEMP=Temp_vary.pos.in.Eval.temp[nb.of.time]
      }
  
  Varying.temp.lengths.progression<-vector("list",length(observed.lengths))
  for (s in 1:length(observed.lengths)){ 
    Varying.temp.lengths.progression[[s]] = growth_process(GrowthCurves=GrowthCurves[[s]], EvalPoints=EvalPoints[[s]], Eval.temp, temp.series=POS.OF.BACK.TEMP, nb.of.time=nb.of.time, max.times =Locating.maxima.benchmark.growthcurves[[s]])
  }
  ### time progression ###
  Times.progression = 0:(nb.of.time - 1)
  ### estimated post-mortem interval ###
  pmi = nb.of.time - 1
  return(list(Growth.shapes = Varying.temp.lengths.progression, Times.progression = Times.progression, start.opt = start.opt, pmi = pmi, Criterion=Criterion))
}





growth_process = function(GrowthCurves, EvalPoints, Eval.temp, temp.series, nb.of.time, max.times,postfeeding=FALSE)
  ######################################################################################################
# Estimation of the temperature-dependent growth profile
#######################################################################################################
# "GrowthCurves" list containing the constant-temperature growth curves
# "EvalPoints" list containing the corresponding sorted grid points 
# "Eval.temp" vector of temperatures.
# "nb.of.time" vector of time points in which the process is going to be evaluated
# "max.times" vector of times at which the costant-temperature curves reach maximum.
# "postfeeding" logical, vector which denotes if each species was observed in the postfeeding phase. NULL if this information is not available.
##################################################################################################
## Returns a list containing
# "$Length.progression" growth curve.
# "$postfeeding" logical, if the estimated process reached postfeeding phase.
##################################################################################################
{
  Length.progression = sapply(GrowthCurves, function(x) x[1])[temp.series[1]]
  increasing<-TRUE
  K = length(GrowthCurves)

  for(kk in 1:K){
    GrowthCurves[[kk]] = GrowthCurves[[kk]] - GrowthCurves[[kk]][1] + Length.progression[1]
    
  }
  
  Updated.delta.length = 0
  for(ll in 2:nb.of.time){
    if(!is.na(Length.progression[ll-1])){ 
   
      kk<- temp.series[ll]
      tm<-mean(Eval.temp[temp.series[1:ll]])
      t0<-which(Eval.temp>tm)[1]-1
      if (max.times[kk]==1){
        Updated.delta.length = diff(approx(EvalPoints[[kk]], GrowthCurves[[kk]], xout = c(1, 2), rule = 1)$y)	
      }else{  
        if(increasing || max(Length.progression)<max(GrowthCurves[[t0]])){ 
       Eval = EvalPoints[[kk]][1:max.times[kk]]
       Growth = GrowthCurves[[kk]][1:max.times[kk]]
       }else{
       Eval = EvalPoints[[kk]][max.times[kk]:length(EvalPoints[[kk]])]
       Growth = GrowthCurves[[kk]][max.times[kk]:length(GrowthCurves[[kk]])]
       }
       updated.time = approx(Growth, Eval, xout = Length.progression[ll - 1], rule = 2)$y
       Updated.delta.length = diff(approx(EvalPoints[[kk]], GrowthCurves[[kk]], xout = c(updated.time, updated.time + 1), rule = 1)$y)
      }
     
    Length.progression[ll] = Length.progression[ll - 1] + Updated.delta.length
    }else{
      Length.progression[ll]<-NA
    }
    if(!is.na(Updated.delta.length) & Updated.delta.length<0){
      increasing<-FALSE
    }
  }
  out<-list(Length.progression=Length.progression,postfeeding=!(increasing))
  return(out)
}





constant_temp_growth= function(observed.lengths, Times_temp_vary, Temp_vary,res.agpaact, Eval.temp)

{
  ######################################################################################################
  # Reconstruction of the constant temperature growth curves for all the temp in the time interval
  #######################################################################################################
  # "observed.lengths" vector containing observed lengths of maggots taken into the corpse
  # "Times_temp_vary" vector of times at which the varying temperatures are observed/estimated. 
  # "Temp_vary" vector of observed/estimated temperature at times "Times_temp_vary". 
  # "res.agpaact" output of the approx.GP.reg function, containing constant-temperature growth curves for all
  #               temperature in Eval.temp
  # "Eval.temp" vector of temperatures
  ##################################################################################################
  # Returns a list containing
  # "$GrowthCurves" estimated growth curve
  # "$EvalPoints" vector containing the corresponding time progression
  ##################################################################################################
  nb.temp_vary = length(Temp_vary)
  Temp_vary.pos.in.Eval.temp = as.numeric(factor(Temp_vary, levels = Eval.temp, labels = 1:length(Eval.temp)))
  Locating.maxima.benchmark.growthcurves = vector("list",length(observed.lengths))
  EvalPoints = vector("list",length(observed.lengths))
  GrowthCurves = vector("list",length(observed.lengths))
  
  for (s in 1:length(observed.lengths)){ 
    Locating.maxima.benchmark.growthcurves[[s]] = sapply(res.agpaact[[s]]$approx.GP, which.max)
    K = length(res.agpaact[[s]]$approx.GP)
    for(kk in 1:K){
      EvalPoints[[s]][[kk]] = res.agpaact[[s]]$approx.eval[[kk]]
      GrowthCurves[[s]][[kk]] = res.agpaact[[s]]$approx.GP[[kk]]
    }
  }
  
  
  return(list(GrowthCurves = GrowthCurves, EvalPoints=EvalPoints))
}





growth_shape= function(observed.lengths, Times_temp_vary, Temp_vary,res.agpaact, Eval.temp,  start=1,postfeeding=FALSE)

{
  ######################################################################################################
  # Estimation of the growth shape given a starting point
  #######################################################################################################
  ######################################################################################################
  # Reconstruction of the constant temperature growth curves for all the temp in the time interval
  #######################################################################################################
  # "observed.lengths" vector containing observed lengths of maggots taken into the corpse
  # "Times_temp_vary" vector of times at which the varying temperatures are observed/estimated. 
  # "Temp_vary" vector of observed/estimated temperature at times "Times_temp_vary". 
  # "res.agpaact" output of the approx.GP.reg function, containing constant-temperature growth curves for all
  #               temperature in Eval.temp
  # "Eval.temp" vector of temperatures
  # "Start" starting point
  ##################################################################################################
  # Returns a list containing
  # "$Growth.shapes" list containing the estimated temperature-dependent growth curves
  # "$Times.progression" list containing the corresponding time progression
  ##################################################################################################
  nb.temp_vary = length(Temp_vary)
  Temp_vary.pos.in.Eval.temp = as.numeric(factor(Temp_vary, levels = Eval.temp, labels = 1:length(Eval.temp)))
  Locating.maxima.benchmark.growthcurves = vector("list",length(observed.lengths))
  EvalPoints = vector("list",length(observed.lengths))
  GrowthCurves = vector("list",length(observed.lengths))
  
  for (s in 1:length(observed.lengths)){ 
    Locating.maxima.benchmark.growthcurves[[s]] = sapply(res.agpaact[[s]]$approx.GP, which.max)
    K = length(res.agpaact[[s]]$approx.GP)
    for(kk in 1:K){
      EvalPoints[[s]][[kk]] = res.agpaact[[s]]$approx.eval[[kk]]
      GrowthCurves[[s]][[kk]] = res.agpaact[[s]]$approx.GP[[kk]]
    }
  }
  
  if(start>=1 & start<=nb.temp_vary){ 
  start.opt = start
  }else if(start<1){
    start.opt<-1
  }else{
    start.opt<-nb.temp_vary
    }
  
  ### estimating the shape of the temperature-dependent growth curves
  nb.of.time = nb.temp_vary - start.opt + 1
  if(start.opt>1 & start.opt<nb.temp_vary){ 
      POS.OF.BACK.TEMP = Temp_vary.pos.in.Eval.temp[-(1:(start.opt-1))]
      }else if(start==1){
        POS.OF.BACK.TEMP = Temp_vary.pos.in.Eval.temp
       
      }else{
       POS.OF.BACK.TEMP=Temp_vary.pos.in.Eval.temp[nb.of.time]
      }
  
  Varying.temp.lengths.progression<-vector("list",length(observed.lengths))
  for (s in 1:length(observed.lengths)){ 
    Varying.temp.lengths.progression[[s]] = growth_process(GrowthCurves=GrowthCurves[[s]], EvalPoints=EvalPoints[[s]], Eval.temp, temp.series=POS.OF.BACK.TEMP, nb.of.time=nb.of.time, max.times =Locating.maxima.benchmark.growthcurves[[s]],postfeeding = postfeeding)
  }
  ### time progression ###
  Times.progression = 0:(nb.of.time - 1)

  return(list(Growth.shapes = Varying.temp.lengths.progression, Times.progression = Times.progression))
}






